import arrow
from linebot import LineBotApi
from linebot.models import (
    TemplateSendMessage,
    ButtonsTemplate,
    MessageAction,
    URIAction
)
from eyewitness.config import BBOX, LINE_FALSE_ALERT_MSG_TEMPLATE, LINE_PLATFROM
from eyewitness.detection_utils import DetectionResultHandler
from eyewitness.models.feedback_models import RegisteredAudience
from eyewitness.models.db_proxy import DATABASE_PROXY


class LineAnnotationSender(DetectionResultHandler):
    def __init__(self, channel_access_token, image_url_handler, audience_ids=None,
                 update_audience_period=0, detection_result_filter=None, detection_method=BBOX,
                 database=None):
        """
        Parameters
        ----------
        audience_ids: set
            line audience ids
        channel_access_token: str
            channel_access_token
        image_url_handler: Callable
            compose drawn image to image_url
        update_audience_period: int
            the period(seconds) that update the audience_ids from audience model, 0 will not update
        detection_result_filter: Callable
            a function check if the detection result to sent or not
        detection_method: str
            detection method
        database: peewee.Database
            peewee database obj, used to query registered audiences
        """
        self.water_mark_time = arrow.now()
        self.update_audience_period = update_audience_period
        self.line_bot_api = LineBotApi(channel_access_token)
        self.image_url_handler = image_url_handler
        self.detection_result_filter = detection_result_filter
        self._detection_method = detection_method
        self.database = database
        self.create_db_table()
        if audience_ids is None:
            self.audience_ids = self.get_registered_audiences()
        else:
            self.audience_ids = audience_ids

    @property
    def detection_method(self):
        """str: detection_method"""
        return self._detection_method

    def create_db_table(self):
        """create the RegisteredAudience if not exist
        """
        self.check_proxy_db()
        RegisteredAudience.create_table()

    def get_registered_audiences(self):
        """ get the RegisteredAudience id list
        """
        if self.database is None:
            raise("the database is not set")
        self.check_proxy_db()
        query = RegisteredAudience.select().where(RegisteredAudience.platform_id == LINE_PLATFROM)
        audiences = set(i.user_id for i in query)
        return audiences

    def check_proxy_db(self):
        """check if the db proxy is correct one, if not initialize again.
        """
        if not (self.database is DATABASE_PROXY.obj):
            DATABASE_PROXY.initialize(self.database)

    def audience_update(self):
        """
        update the audiences from RegisteredAudience model
        """
        if self.update_audience_period and self.database is not None:
            diff_seconds = (arrow.now() - self.water_mark_time).total_seconds()
            if diff_seconds > self.update_audience_period:
                audience_ids = self.get_registered_audiences()
                self.audience_ids = audience_ids

    def _handle(self, detection_result):
        # update audience
        self.audience_update()

        # check if detection result need to sent_out
        if self.detection_result_filter(detection_result):
            image_url = self.image_url_handler(detection_result.drawn_image_path)
            false_alert_feedback_text = LINE_FALSE_ALERT_MSG_TEMPLATE.format(
                image_id=str(detection_result.image_id), meta='')
            self.send_annotation_button_msg(image_url, false_alert_feedback_text)

    def send_annotation_button_msg(self, image_url, false_alert_feedback_text):
        """
        sent line botton msg to audience_ids

        Parameters
        ----------
        image_url: str
            the url of image

        false_alert_feedback_text: str
            false_alert msg used to sent to feedback_handler
        """
        buttons_msg = TemplateSendMessage(
            alt_text='object detected',
            template=ButtonsTemplate(
                thumbnail_image_url=image_url,
                title='object detected',
                text='help to report result',
                actions=[
                    MessageAction(
                        label='Report Error (錯誤回報)',
                        text=false_alert_feedback_text
                    ),
                    URIAction(
                        label='full image (完整圖片)',
                        uri=image_url
                    )
                ]
            )
        )
        if self.audience_ids:  # check if audiences
            self.line_bot_api.multicast(list(self.audience_ids), buttons_msg)



import six
from abc import ABCMeta, abstractmethod


@six.add_metaclass(ABCMeta)
class ObjectDetector():
    """
    Abstract class used to wrapper object detector
    """
    @abstractmethod
    def detect(self, image, image_id):
        """

        [abstract method] need to implement detection method which return DetectionResult obj

        Parameters
        ----------
        image: PIL.Image
            PIL.Image instance
        image_id: ImageId
            image_id obj

        Returns
        -------
        DetectionResult: DetectionResult
            the detected result of given image
        """
        pass



import json
import six
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from eyewitness.config import (
    FEEDBACK_BBOX,
    FEEDBACK_FALSE_ALERT,
    IMAGE_ID,
    AUDIENCE_ID,
    FEEDBACK_METHOD,
    FEEDBACK_MSG_OBJS,
    RECEIVE_TIME,
    FEEDBACK_META,
)
from eyewitness.image_id import ImageId
from eyewitness.audience_id import AudienceId


BboxObjectFeedback = namedtuple('BboxObjectFeedback',
                                ['x1', 'y1', 'x2', 'y2', 'label', 'meta'])
Type_Serialization_Mapping = {
    FEEDBACK_BBOX: BboxObjectFeedback,
    FEEDBACK_FALSE_ALERT: None,  # false alert won't have feedback_msg_objs
}


class FeedbackMsg(object):
    """
    represent the Feedback msg
    """
    def __init__(self, feedback_dict):
        """
        Parameters
        -----------
        feedback_dict: dict
            - audience_id: AudienceId
                the audience who feedback the msg
            - feedback_method: str
                which kind of feedback
            - image_id: ImageId
                the ImageId related to feedback
            - feedback_meta: str
                misc feedback msg
            - feedback_msg_objs: List[tuple]
                feedback objs (e.g. bboxs)
            - receive_time: int
                the timestamp receive the msg
        """
        # get feedback method
        self.feedback_method = feedback_dict.get(FEEDBACK_METHOD, FEEDBACK_FALSE_ALERT)
        self.feedback_type = Type_Serialization_Mapping[self.feedback_method]
        self.feedback_dict = feedback_dict

    @property
    def image_id(self):
        """ImageId: ImageId obj"""
        return self.feedback_dict.get(IMAGE_ID)

    @property
    def audience_id(self):
        """AudienceId: AudienceId obj"""
        return self.feedback_dict[AUDIENCE_ID]

    @property
    def receive_time(self):
        """int: received timestamp"""
        return self.feedback_dict[RECEIVE_TIME]

    @property
    def feedback_msg_objs(self):
        """List[tuples]: List of msg named tuple objs"""
        feedback_msg_objs = self.feedback_dict.get(FEEDBACK_MSG_OBJS, [])
        return [self.feedback_type(*i) for i in feedback_msg_objs]

    @property
    def feedback_meta(self):
        """str: feedback_meta str"""
        return self.feedback_dict.get(FEEDBACK_META)

    @property
    def is_false_alert(self):
        """bool: is false_alert or not"""
        return self.feedback_dict.get('is_false_alert', True)

    @classmethod
    def from_json(cls, json_str):
        """
        Parameters
        ----------
        json_str: str
            feedback_msg json str

        Returns
        -------
        feedback_msg_obj: FeedbackMsg
            a feedback msg instance
        """
        feedback_json_dict = json.loads(json_str)
        feedback_json_dict[AUDIENCE_ID] = AudienceId.from_str(feedback_json_dict[AUDIENCE_ID])
        if feedback_json_dict[IMAGE_ID]:
            feedback_json_dict[IMAGE_ID] = ImageId.from_str(feedback_json_dict[IMAGE_ID])
        return cls(feedback_json_dict)

    def to_json_dict(self):
        """
        Returns
        -------
        image_dict: dict
            the dict repsentation of detection_result
        """
        json_dict = dict(self.feedback_json_dict)
        json_dict[AUDIENCE_ID] = str(json_dict[AUDIENCE_ID])
        if json_dict[IMAGE_ID]:
            json_dict[IMAGE_ID] = str(json_dict[IMAGE_ID])
        return json_dict


@six.add_metaclass(ABCMeta)
class FeedbackMsgHandler():
    """
    Abstract class for FeedbackMsgHandler

    with a abstract method *_handle(feedback_msg)* used to handle feedback msg
    """
    def handle(self, feedback_msg):
        """
        a wrapper for *_handle(feedback_msg)* and *feedback_method* check
        """
        assert self.feedback_method == feedback_msg.feedback_method
        return self._handle(feedback_msg)

    @abstractmethod
    def _handle(self, feedback_msg):
        """
        abstract method for handle feedback_msg

        Parameters
        ----------
        feedback_msg: FeedbackMsg
        """
        pass

    @property
    def feedback_method(self):
        """str: feedback_method"""
        raise NotImplementedError



from peewee import (
    BooleanField,
    CharField,
    TimestampField,
    ForeignKeyField,
    IntegerField,
    Model
)

from eyewitness.models.db_proxy import DATABASE_PROXY
from eyewitness.models.detection_models import ImageInfo


class BaseModel(Model):
    class Meta:
        database = DATABASE_PROXY


class RegisteredAudience(BaseModel):
    audience_id = CharField(unique=True, primary_key=True)
    user_id = CharField(null=False)
    platform_id = CharField(null=False)
    register_time = TimestampField()
    description = CharField()


class FalseAlertFeedback(BaseModel):
    # peewee didn't support compositeKey as foreignKey, using field to specify field
    audience_id = ForeignKeyField(RegisteredAudience)
    image_id = ForeignKeyField(ImageInfo, null=True)
    receive_time = TimestampField()
    feedback_meta = CharField()
    # TODO: if the is_false_alert field needed??
    is_false_alert = BooleanField()


class BboxAnnotationFeedback(BaseModel):
    # peewee didn't support compositeKey as foreignKey, using field to specify field
    audience_id = ForeignKeyField(RegisteredAudience)
    image_id = ForeignKeyField(ImageInfo, null=True)
    receive_time = TimestampField()
    feedback_meta = CharField()
    is_false_alert = BooleanField()
    x1 = IntegerField()
    x2 = IntegerField()
    y1 = IntegerField()
    y2 = IntegerField()
    label = CharField()



# detection type
BBOX = 'bbox'

# Feedback type
FEEDBACK_BBOX = 'feedback_bbox'
FEEDBACK_FALSE_ALERT = 'feedback_false_alert'

# image producer type
POST_PATH = 'post_path'
POST_BYTES = 'post_bytes'
IN_MEMORY = 'in_memory'

# Line false_alert msg template
LINE_FALSE_ALERT_MSG_TEMPLATE = "false_alert_{image_id}_{meta}"
LINE_FALSE_ALERT_ANNOTATION_PATTERN = (
    r"false_alert_(?P<channel>.*)::(?P<timestamp>\d*)::(?P<file_format>.*)_(?P<meta>.*)")

# audience_platfrom
LINE_PLATFROM = 'line'

# image_info fields
RAW_IMAGE_PATH = 'raw_image_path'
IMAGE_ID = 'image_id'

# detected_result fields
DRAWN_IMAGE_PATH = 'drawn_image_path'
DETECTED_OBJECTS = 'detected_objects'
DETECTION_METHOD = 'detection_method'

# audience_info fields
AUDIENCE_ID = 'audience_id'
RECEIVE_TIME = 'receive_time'

# feedback infomation
FEEDBACK_METHOD = 'feedback_method'
FEEDBACK_MSG_OBJS = 'feedback_msg_objs'
FEEDBACK_META = 'feedback_meta'
IS_FALSE_ALERT = 'is_false_alert'



import json
import six
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from eyewitness.config import (
    BBOX,
    DRAWN_IMAGE_PATH,
    IMAGE_ID,
    DETECTED_OBJECTS,
    DETECTION_METHOD
)
from eyewitness.image_id import ImageId


BoundedBoxObject = namedtuple(
    'BoundedBoxObject', ['x1', 'y1', 'x2', 'y2', 'label', 'score', 'meta'])
Type_Serialization_Mapping = {BBOX: BoundedBoxObject}


class DetectionResult(object):
    """
    represent detection result of a image.
    """

    def __init__(self, image_dict):
        """
        Parameters
        ----------
        image_dict: dict
            - detection_method: detection_method str

            - detected_objects: List[tuple], list of detected obj (optional)

            - drawn_image_path: str, path of drawn image (optional)

            - image_id: image_id obj
        """
        # get detection method
        self.detection_method = image_dict.get(DETECTION_METHOD, BBOX)
        self.detection_type = Type_Serialization_Mapping[self.detection_method]
        self.image_dict = image_dict

    @property
    def image_id(self):
        """ImageId: image_id obj"""
        return self.image_dict[IMAGE_ID]

    @property
    def drawn_image_path(self):
        """str: drawn_image_path"""
        return self.image_dict.get(DRAWN_IMAGE_PATH, '')

    @property
    def detected_objects(self):
        """List[object]: List of detected objects in the image"""
        detected_objects = self.image_dict.get(DETECTED_OBJECTS, [])
        return [self.detection_type(*i) for i in detected_objects]

    @classmethod
    def from_json(cls, json_str):
        img_json_dict = json.loads(json_str)
        img_json_dict[IMAGE_ID] = ImageId.from_str(img_json_dict[IMAGE_ID])
        return cls(img_json_dict)

    def to_json_dict(self):
        """
        Returns
        -------
        image_dict: dict
            the dict repsentation of detection_result
        """
        json_dict = dict(self.image_dict)
        json_dict[IMAGE_ID] = str(json_dict[IMAGE_ID])
        return json_dict


@six.add_metaclass(ABCMeta)
class DetectionResultHandler():
    """a abstract class design to handle detection result
    need to implement:

    - function: _handle(self, detection_result)

    - property: detection_method
    """
    @abstractmethod
    def _handle(self, detection_result):
        """abstract method for handle DetectionResult

        Parameters
        ----------
        detection_result: DetectionResult
        """
        pass

    def handle(self, detection_result):
        """wrapper of _handle function with the check of detection_method with detection_result.

        Parameters
        ----------
        detection_result: DetectionResult
        """
        assert self.detection_method == detection_result.detection_method
        self._handle(detection_result)

    @property
    def detection_method(self):
        raise NotImplementedError



from setuptools import setup, find_packages


description = """\
A light weight framework for Object Detection."""

install_requires = [
    'nose2',
    'Pillow==5.2.0',
    'line-bot-sdk',
    'flask==1.0.2',
    'arrow==0.10.0',
    'six',
    'peewee==3.7.1',
    'wtf-peewee==3.0.0',
    'flask-admin==1.5.2',
    'lxml==4.2.4'
]


setup(
    name='eyewitness',
    version='0.58',
    description=description,
    author='Ching-Hua Yang',
    url='https://gitlab.com/penolove15/witness',
    install_requires=install_requires,
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(),
    include_package_data=True
)



from __future__ import print_function
import os
import glob
import random
from shutil import copyfile

from lxml import etree
from eyewitness.utils import make_path
from eyewitness.models.db_proxy import DATABASE_PROXY
from eyewitness.models.feedback_models import FalseAlertFeedback
from eyewitness.models.detection_models import BboxDetectionResult


def add_filename_prefix(filename, prefix):
    return "%s_%s" % (prefix, filename)


def create_bbox_dataset_from_eyewitness(
        database, valid_classes, output_dataset_folder, dataset_name):
    """
    generate bbox dataset from eyewitness

    - remove images with false-alert feedback

    - get images with selected classes objects
    """
    anno_folder = os.path.join(output_dataset_folder, 'Annotations')
    jpg_images_folder = os.path.join(output_dataset_folder, 'JPEGImages')
    main_folder = os.path.join(output_dataset_folder, 'ImageSets', 'Main')

    # mkdir if not exist
    make_path(anno_folder)
    make_path(jpg_images_folder)
    make_path(main_folder)

    DATABASE_PROXY.initialize(database)
    # filter false alert, and valid_classes
    false_alert_query = FalseAlertFeedback.select(
        FalseAlertFeedback.image_id).where(FalseAlertFeedback.is_false_alert)
    valid_objects = BboxDetectionResult.select().where(
        BboxDetectionResult.image_id.not_in(false_alert_query),
        BboxDetectionResult.label.in_(valid_classes))

    # get valid_images with raw_image_path
    valid_images = set(i.image_id for i in valid_objects if i.image_id.raw_image_path)

    # generate etree obj for each images
    for valid_image in valid_images:
        if (valid_image.file_format != 'jpg' or not os.path.exists(valid_image.raw_image_path)):
            # TODO: support other file_format
            continue

        ori_image_file = valid_image.raw_image_path
        dest_image_file = os.path.join(jpg_images_folder, "%s.jpg" % valid_image.image_id)
        copyfile(ori_image_file, dest_image_file)

        # prepare anno_file
        anno_file = os.path.join(anno_folder, "%s.xml" % valid_image.image_id)
        detected_objects = list(BboxDetectionResult.select().where(
            BboxDetectionResult.image_id == valid_image,
            BboxDetectionResult.label.in_(valid_classes)))
        if detected_objects:  # make sure there is detected objects
            etree_obj = generate_etree_obj(valid_image, detected_objects, dataset_name)
            etree_obj.write(anno_file, pretty_print=True)


def generate_etree_obj(valid_image, detected_objects, dataset_name):
    root = etree.Element("annotation")
    filename = etree.SubElement(root, "filename")
    source = etree.SubElement(root, "source")
    databases = etree.SubElement(source, "database")

    filename.text = valid_image.image_id
    databases.text = dataset_name
    for obj in detected_objects:
        object_ = etree.SubElement(root, "object")
        name = etree.SubElement(object_, "name")
        name.text = obj.label
        pose = etree.SubElement(object_, "pose")
        pose.text = "Unspecified"
        truncated = etree.SubElement(object_, "truncated")
        truncated.text = "0"
        difficult = etree.SubElement(object_, "difficult")
        difficult.text = "0"
        # bounded box
        bndbox = etree.SubElement(object_, "bndbox")
        xmin_ = etree.SubElement(bndbox, "xmin")
        ymin_ = etree.SubElement(bndbox, "ymin")
        xmax_ = etree.SubElement(bndbox, "xmax")
        ymax_ = etree.SubElement(bndbox, "ymax")
        xmin_.text = str(obj.x1)
        ymin_.text = str(obj.y1)
        xmax_.text = str(obj.x2)
        ymax_.text = str(obj.y2)
    return etree.ElementTree(root)


def copy_image_to_output_dataset(filename, src_dataset, jpg_images_folder, anno_folder, file_fp):
    filename_with_prefix = add_filename_prefix(filename, src_dataset.dataset_name)

    # copy image file
    ori_image_file = os.path.join(src_dataset.jpg_images_folder, "%s.jpg" % filename)
    dest_image_file = os.path.join(jpg_images_folder, "%s.jpg" % filename_with_prefix)
    copyfile(ori_image_file, dest_image_file)

    # copy annotation file
    ori_anno_file = os.path.join(src_dataset.anno_folder, "%s.xml" % filename)
    dest_anno_file = os.path.join(anno_folder, "%s.xml" % filename_with_prefix)
    copyfile(ori_anno_file, dest_anno_file)

    # print filename to the filename list file
    print(filename_with_prefix, file=file_fp)


class BboxDataSet(object):
    """
    generate DataSet with same format as VOC object detections:

    <dataset_folder>/Annotations/<image_name>.xml

    <dataset_folder>/JPEGImages/<image_name>.jpg

    <dataset_folder>/ImageSets/Main/trainval.txt

    <dataset_folder>/ImageSets/Main/test.txt

    """
    def __init__(self, dataset_folder, dataset_name):
        self.anno_folder = os.path.join(dataset_folder, 'Annotations')
        self.jpg_images_folder = os.path.join(dataset_folder, 'JPEGImages')
        self.main_folder = os.path.join(dataset_folder, 'ImageSets', 'Main')
        self.trainval_file = os.path.join(self.main_folder, 'trainval.txt')
        self.test_file = os.path.join(self.main_folder, 'test.txt')
        self.dataset_name = dataset_name

    @property
    def training_and_validation_set(self):
        with open(self.trainval_file) as f:
            for i in f:
                yield i.strip()

    @property
    def testing_set(self):
        with open(self.test_file) as f:
            for i in f:
                yield i.strip()

    def generate_train_test_list(self, overwrite=True, train_ratio=0.9):
        if not overwrite and os.path.exists(self.trainval_file) and os.path.exists(self.test_file):
            return
        else:
            anno_files = glob.glob('%s/*.xml' % self.anno_folder)
            image_ids_anno = set(
                anno_file.rsplit('/', 1)[1].replace('.xml', '') for anno_file in anno_files)
            jpg_files = glob.glob('%s/*.jpg' % self.jpg_images_folder)
            image_ids_jpg = set(
                jpg_file.rsplit('/', 1)[1].replace('.jpg', '') for jpg_file in jpg_files)
            image_ids = image_ids_jpg.intersection(image_ids_anno)

            # write to training set
            training_set = random.sample(image_ids, int(len(image_ids) * train_ratio))
            with open(self.trainval_file, 'w') as f:
                for train_id in training_set:
                    print(train_id, file=f)

            testing_set = image_ids.difference(training_set)
            with open(self.test_file, 'w') as f:
                for test_id in testing_set:
                    print(test_id, file=f)

    @classmethod
    def union_bbox_datasets(cls, datasets, output_dataset_folder, dataset_name):
        """
        union bbox datasets and copy files to the given output_dataset
        """
        anno_folder = os.path.join(output_dataset_folder, 'Annotations')
        jpg_images_folder = os.path.join(output_dataset_folder, 'JPEGImages')
        main_folder = os.path.join(output_dataset_folder, 'ImageSets', 'Main')

        # mkdir if not exist
        make_path(anno_folder)
        make_path(jpg_images_folder)
        make_path(main_folder)

        # write train, test list out
        trainval_file = os.path.join(main_folder, 'trainval.txt')
        test_file = os.path.join(main_folder, 'test.txt')
        with open(trainval_file, 'w') as train_fp, open(test_file, 'w') as test_fp:
            for dataset in datasets:
                for train_file in dataset.training_and_validation_set:
                    copy_image_to_output_dataset(
                        train_file, dataset, jpg_images_folder, anno_folder, train_fp)

                for test_file in dataset.testing_set:
                    copy_image_to_output_dataset(
                        test_file, dataset, jpg_images_folder, anno_folder, test_fp)

        return cls(output_dataset_folder, dataset_name)



import six
from abc import ABCMeta, abstractmethod
from eyewitness.config import RAW_IMAGE_PATH


class ImageId(object):
    """
    ImageId is used to standardize the image_id format
    """

    def __init__(self, channel, timestamp, file_format='jpg'):
        """
        Parameters
        ----------
        channel: str
            channel of image comes
        timestamp: int
            timestamp of image arrive time
        format: str
            type of image
        """
        self.timestamp = timestamp
        self.channel = channel
        self.file_format = file_format

    def __hash__(self):
        return hash(self.timestamp) ^ hash(self.channel) ^ hash(self.file_format)

    def __str__(self):
        # TODO make a more safer way
        return "{}::{}::{}".format(self.channel, str(self.timestamp), self.file_format)

    def __eq__(self, other):
        if isinstance(other, ImageId):
            return str(self) == str(other)
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, ImageId):
            return str(self) < str(other)
        return NotImplemented

    @classmethod
    def from_str(cls, image_id_str):
        """
        Parameters
        ----------
        image_id_str: str
            a string with pattern {chanel}::{timestamp}::{fileformat}
            e.g: "channel::12345567::jpg"

        Returns
        -------
        image_id: ImageId
            a ImageId obj
        """
        # TODO make a more safer way
        channel, timestamp, file_format = image_id_str.split('::')
        return cls(channel=channel, timestamp=int(timestamp), file_format=file_format)


@six.add_metaclass(ABCMeta)
class ImageRegister():
    def register_image(self, image_id, meta_dict):
        """
        interface for ImageRegister to register_image
        """
        raw_image_path = meta_dict.get(RAW_IMAGE_PATH, None)
        self.insert_image_info(image_id, raw_image_path=raw_image_path)

    @abstractmethod
    def insert_image_info(image_id, raw_image_path):
        """
        abstract method which need to be implement: how to insert/record image information

        Parameters
        ----------
        image_id: ImageId
            ImageId obj
        raw_image_path: str
            the path of raw image
        """
        pass



import sqlite3

from eyewitness.config import BBOX, FEEDBACK_FALSE_ALERT
from eyewitness.detection_utils import DetectionResultHandler
from eyewitness.audience_id import AudienceRegister
from eyewitness.feedback_msg_utils import FeedbackMsgHandler
from eyewitness.image_id import ImageRegister
from eyewitness.models import detection_models
from eyewitness.models import feedback_models
from eyewitness.models.db_proxy import DATABASE_PROXY


class BboxNativeSQLiteDbWriter(DetectionResultHandler, ImageRegister):
    def __init__(self, db_path):
        """
        Parameters
        ----------
        db_path: str
            database path
        """
        self.conn = sqlite3.connect(db_path)
        self.create_db_table()

    @property
    def detection_method(self):
        """str: BBOX"""
        return BBOX

    def _handle(self, detection_result):
        """
        handle detection result

        Parameters
        ----------
        detection_result: DetectionResult
            detection result
        """
        image_id = detection_result.image_id
        self.insert_detection_objs(
            str(image_id), detection_result.detected_objects)
        if detection_result.drawn_image_path:
            self.update_image_drawn_image_path(image_id, detection_result.drawn_image_path)

    def create_db_table(self):
        """
        create ImageInfo, BboxDetectionResult table if table not exist
        """
        print('connet/create image_info table')

        self.conn.execute('''CREATE TABLE IF NOT EXISTS ImageInfo(
                          image_id TEXT PRIMARY KEY,
                          timestamp INTEGER,
                          channel TEXT,
                          file_format TEXT,
                          raw_image_path TEXT,
                          drawn_image_path TEXT
                          )''')

        print('connet/create bbox_detection_results table')
        self.conn.execute('''CREATE TABLE IF NOT EXISTS BboxDetectionResult(
                          ID INTEGER PRIMARY KEY AUTOINCREMENT,
                          image_id TEXT,
                          x1 INTEGER,
                          y1 INTEGER,
                          x2 INTEGER,
                          y2 INTEGER,
                          label TEXT,
                          score REAL,
                          meta TEXT,
                          FOREIGN KEY(image_id) REFERENCES image_info(image_id)
                          )''')
        self.conn.commit()

    def insert_image_info(self, image_id, raw_image_path=None):
        """
        insert image_info which used for unit-test

        Parameters
        ----------
        image_id: str
            image_id
        raw_image_path: str
             the path of raw image stored
        """
        try:
            timestamp = image_id.timestamp
            channel = image_id.channel
            file_format = image_id.file_format
            self.conn.execute('''INSERT INTO ImageInfo(image_id, timestamp, channel, file_format,
                              raw_image_path) VALUES(?,?,?,?,?)''',
                              (str(image_id), timestamp, channel, file_format, raw_image_path))
            self.conn.commit()
        except Exception:
            pass

    def update_image_drawn_image_path(self, image_id, drawn_image_path):
        """
        update db image_id.drawn_image_path
        """
        try:
            self.conn.execute('''UPDATE ImageInfo SET drawn_image_path = ? where image_id = ? ''',
                              (drawn_image_path, str(image_id)))
            self.conn.commit()
        except ValueError:
            print("update failure")
            pass

    def insert_detection_objs(self, image_id, detected_objects):
        """
        insert detection results into db.

        Parameters
        ----------
        image_id: str
            image_id
        detected_objects: List[BoundedBoxObject]
            detected objects
        """
        try:
            for detected_obj in detected_objects:
                left, top, right, bottom, label, score, meta = detected_obj
                self.conn.execute(
                    '''INSERT INTO BboxDetectionResult
                       (image_id, x1, y1, x2, y2, label, score, meta)
                        VALUES (?, ?, ?, ?, ?, ?, ?, ?)''',
                    (image_id, left, top, right, bottom, label, score, meta))
            self.conn.commit()
        except sqlite3.IntegrityError:
            pass


class BboxPeeweeDbWriter(DetectionResultHandler, ImageRegister):
    def __init__(self, database):
        """
        Parameters
        ----------
        database: peewee.Database
            peewee db obj
        """
        # setup database for models
        self.database = database
        DATABASE_PROXY.initialize(self.database)
        self.create_db_table()

    @property
    def detection_method(self):
        return BBOX

    def check_proxy_db(self):
        """check if the db proxy is correct one, if not initialize again.
        """
        if not (self.database is DATABASE_PROXY.obj):
            DATABASE_PROXY.initialize(self.database)

    def _handle(self, detection_result):
        """
        Parameters
        ----------
        detection_result: DetectionResult
            detection_result
        """
        image_id = detection_result.image_id
        self.insert_detection_objs(
            image_id, detection_result.detected_objects)
        if detection_result.drawn_image_path:
            self.update_image_drawn_image_path(image_id, detection_result.drawn_image_path)

    def create_db_table(self):
        """
        create ImageInfo, BboxDetectionResult table if table not exist
        """
        detection_models.ImageInfo.create_table()
        detection_models.BboxDetectionResult.create_table()

    def insert_image_info(self, image_id, raw_image_path=None):
        """
        insert image_info which used for unit-test

        Parameters
        ----------
        image_id: ImageId obj
            image_id obj (including  channel, timestamp, file-format)
        raw_image_path: str
            the path of raw image stored
        """
        try:
            self.check_proxy_db()
            timestamp = image_id.timestamp
            channel = image_id.channel
            file_format = image_id.file_format
            image_info = detection_models.ImageInfo(
                image_id=str(image_id), timestamp=timestamp, channel=channel,
                file_format=file_format, raw_image_path=raw_image_path)
            # according to document:
            # http://docs.peewee-orm.com/en/latest/peewee/models.html#non-integer-primary-keys-composite-keys-and-other-tricks
            # model with Non-integer primary keys need to pass `force_insert=True`
            image_info.save(force_insert=True)
        except ValueError:
            pass

    def update_image_drawn_image_path(self, image_id, drawn_image_path):
        """
        update db image_id.drawn_image_path
        """
        try:
            self.check_proxy_db()
            image_info = detection_models.ImageInfo(
                image_id=str(image_id), drawn_image_path=drawn_image_path)
            image_info.save(only=[detection_models.ImageInfo.drawn_image_path])
        except ValueError:
            pass

    def insert_detection_objs(self, image_id, detected_objects):
        """
        insert detection results into db.

        Parameters
        ----------
        image_id: str
            image_id
        detected_objects: List[BoundedBoxObject]
            detected objects
        """
        try:
            self.check_proxy_db()
            for detected_obj in detected_objects:
                left, top, right, bottom, label, score, meta = detected_obj
                detection_result = detection_models.BboxDetectionResult(
                    image_id=image_id, x1=left, y1=top, x2=right, y2=bottom, label=label,
                    score=score, meta=meta)
                detection_result.save()
        except ValueError:
            pass


class FalseAlertPeeweeDbWriter(FeedbackMsgHandler, AudienceRegister, ImageRegister):
    def __init__(self, database):
        """
        Parameters
        ----------
        database: peewee.Database
            peewee db obj
        """
        # setup database for models
        self.database = database
        DATABASE_PROXY.initialize(self.database)
        self.create_db_table()

    @property
    def feedback_method(self):
        return FEEDBACK_FALSE_ALERT

    def check_proxy_db(self):
        """check if the db proxy is correct one, if not initialize again.
        """
        if not (self.database is DATABASE_PROXY.obj):
            DATABASE_PROXY.initialize(self.database)

    def _handle(self, feedback_msg):
        """
        Parameters
        ----------
        feedback_msg: FeedbackMsg
            feedback_msg
        """
        self.insert_feedback_obj(feedback_msg)

    def create_db_table(self):
        """
        create ImageInfo, RegisteredAudience, FalseAlertFeedback table if table not exist
        """
        detection_models.ImageInfo.create_table()
        feedback_models.RegisteredAudience.create_table()
        feedback_models.FalseAlertFeedback.create_table()

    def insert_image_info(self, image_id, raw_image_path=None):
        """
        insert image_info which used for unit-test

        Parameters
        ----------
        image_id: str
            image_id
        raw_image_path: str
            the path of raw image stored
        """
        try:
            self.check_proxy_db()
            timestamp = image_id.timestamp
            channel = image_id.channel
            file_format = image_id.file_format
            image_info = detection_models.ImageInfo(
                image_id=str(image_id), timestamp=timestamp, channel=channel,
                file_format=file_format, raw_image_path=raw_image_path)
            # according to document:
            # http://docs.peewee-orm.com/en/latest/peewee/models.html#non-integer-primary-keys-composite-keys-and-other-tricks
            # model with Non-integer primary keys need to pass `force_insert=True`
            image_info.save(force_insert=True)
        except ValueError:
            pass

    def insert_registered_user(self, audience_id, register_time, description):
        """
        insert image_info which used for unit-test

        Parameters
        ----------
        audience_id: AudienceId

        register_time: int

        description: str
        """
        try:
            platform_id = audience_id.platform_id
            user_id = audience_id.user_id
            self.check_proxy_db()
            user_to_register = feedback_models.RegisteredAudience(
                audience_id=str(audience_id), user_id=user_id,
                platform_id=platform_id, register_time=register_time,
                description=description)
            # according to document:
            # http://docs.peewee-orm.com/en/latest/peewee/models.html#non-integer-primary-keys-composite-keys-and-other-tricks
            # model with Non-integer primary keys need to pass `force_insert=True`
            user_to_register.save(force_insert=True)
        except ValueError:
            pass

    def insert_feedback_obj(self, feedback_msg):
        """
        insert feedback obj into db.

        Parameters
        ----------
        feedback_msg: FeedbackMsg
        """
        try:
            audience_id = feedback_msg.audience_id
            user_id = audience_id.user_id
            platform_id = audience_id.platform_id
            image_id = feedback_msg.image_id
            receive_time = feedback_msg.receive_time
            feedback_meta = feedback_msg.feedback_meta
            is_false_alert = feedback_msg.is_false_alert

            self.check_proxy_db()
            feedback_model_obj = feedback_models.FalseAlertFeedback(
                audience_id=str(audience_id), user_id=user_id, platform_id=platform_id,
                image_id=str(image_id), receive_time=receive_time, feedback_meta=feedback_meta,
                is_false_alert=is_false_alert)

            feedback_model_obj.save()
        except ValueError:
            pass
